-- policemod.lua: basic play as police script for SR3 by Gumshoe
-- thanks to nclok1405 and IdolNinja for scripting ideas from their lua mods
-- thanks to antonin for his Saints Row 2 'play as a cop' mod which inspired this
-- thanks to Volition for modding support and lua documentation

-- v 0.1 - https://saintsrowmods.com - Sep 15, 2020

-- settings
local PMOD_SET_CAN_GO_ROGUE = true -- whether or not you can go 'rogue' by killing cops/lots of innocents
local PMOD_SET_POINTS_TO_GO_ROGUE = 100 -- number of rogue notoriety points needed to go rogue
local PMOD_SET_TIME_BEFORE_POINTS_DECAY = 10 -- number of seconds until rogue notoriety points start to decay
local PMOD_SET_TANKS_ENABLED = false -- whether or not tanks spawn as backup (doesn't work?)
local PMOD_SET_AIR_ENABLED = true -- whether or not helicopters spawn as backup (doesn't work?)
local PMOD_SET_FORCE_COP_OUTFIT = false -- if true, cop outfit will be worn automatically

-- controls
local PMOD_KEY_MAIN = { key = "CBA_VDC_HORN", img = "{HORN_IMG}" }  -- the key that must be held to perform the 'main' actions (F, by default)
local PMOD_KEY_ACTION = { key = "CBA_GAC_ACTION", img = "{ACTION_IMG}" } -- the main action key used for most things (E, by default)
local PMOD_KEY_UP = { key = "CBA_FIGHT_CLUB_BLOCK", img = "([format][color:saints]UP[/format])" } -- up arrow
local PMOD_KEY_DOWN = { key = "CBA_VDC_HYDRAULICS_TOGGLE", img = "{HYDRAULICS_TOGGLE_IMG}" } -- down arrow
local PMOD_KEY_RIGHT = { key = "CBA_VDC_HYDRAULICS_FULLY_EXTEND_TOGGLE", img = "{HYDRAULICS_FULLY_EXTEND_TOGGLE_IMG}" } -- unused

-- control prompts/flow
local PMOD_HELD_KEY = nil
local PMOD_HELD_KEY_TICK = 0

-- variables
local PMOD_LAST_MSG = 0
local PMOD_IS_A_COP = false
local PMOD_DISPATCH_TICK = 0
local PMOD_HAS_HOSTAGE = 0
local PMOD_CURRENT_HOSTAGE = ""
local PMOD_IS_INTERROGATING = false
local PMOD_SUSPECTS = { }
local PMOD_CURRENT_SUSPECT = ""
local PMOD_CURRENT_NOTORIETY_POINTS = 0
local PMOD_CURRENT_NOTORIETY_DECAY = 0

-- msgs
local PMOD_MSG_BECAME_COP = "[format][color:blue]YOU ARE NOW A COP.[/format]"
local PMOD_MSG_BECAME_SAINT = "[format][color:saints]YOU HAVE BECOME A SAINT AGAIN.[/format]"
local PMOD_MSG_GONE_ROGUE = "YOU HAVE GONE ROGUE!\n[format][color:red]YOU ARE NO LONGER IMMUNE FROM THE LAW.[/format]"
local PMOD_MSG_CONTROLS_SAINT = "PRESS " .. PMOD_KEY_ACTION.img .. " TO BECOME A COP."
local PMOD_MSG_CONTROLS_COP = "PRESS " .. PMOD_KEY_ACTION.img .. " TO STOP BEING A COP.\nUSE " .. PMOD_KEY_UP.img .. " " .. PMOD_KEY_DOWN.img .. " FOR BACKUP."

-- note you could change these to stag or other teams if you really wanted to
local PMOD_TEAM_PLAYER = "Playas"
local PMOD_TEAM_POLICE = "Police"

-- misc
local PMOD_COWER_VARIANTS = { 0, 1, 2, 3, 4 } -- Hunched, Lay Down, Knee, Other Knee, Ball
local PMOD_NOTORIETY_POINTS = 
{
	["carjacking"] = 15,
	["civilian assault"]  = 10,
	["civilian kill"] = 20,
	["police alert"] = 10,
	["police assault"] = 15,
	["police kill"] = 30,
	["atm extortion"] = 50,
	["robbery"] = 50,
	["burglary"] = 50,
	["armored truck assault"] = 50,
	["vandalism"] = 15,
}

-- ** functions start **
function policemod_main()
    debug_print("policemod_main() starting".."\n")

    -- create our threads
    thread_new("policemod_control_thread")
    thread_new("policemod_action_thread")

    -- callback function for when the player takes damage.
    on_take_damage("policemod_on_take_damage", LOCAL_PLAYER)
    -- callback function for when the player's notoriety increases due to a notoriety event.
    on_notoriety_event("policemod_on_notoriety_event", LOCAL_PLAYER)

    debug_print("policemod_main() processed".."\n")
end

-- show messages
function policemod_message(msg)
    if PMOD_LAST_MSG then 
        message_remove(PMOD_LAST_MSG)
    end
    PMOD_LAST_MSG = message(msg, 3.0, false, SYNC_LOCAL)
end

-- change player into police uniform or out of it
function policemod_set_player_clothes(is_a_cop)
    if PMOD_SET_FORCE_COP_OUTFIT then
        if is_a_cop then
            -- give police uniform
            customization_item_wear("cm_policesuit", "cm_policesuit.cmeshx", "mayweather_Body", false, SYNC_ALL)
        else
            -- can't remove or revert clothes through lua so this is the best we can do...
            players_naked()
            if character_get_gender(LOCAL_PLAYER) == 0 then
                customization_item_wear("saints male breifs 1", "cm_udwr_m_briefs01.cmeshx", "01 - Default", false, SYNC_ALL)
            else
                customization_item_wear("saints female bra 1", "cm_bra_f_saints01.cmeshx", "CM_BRA_F_Pink01", false, SYNC_ALL)
                customization_item_wear("saints female panty 1", "cm_udwr_f_saints01.cmeshx", "Panties", false, SYNC_ALL)
            end
        end
        customization_item_revert() -- need to call this to enable shops etc. - it doesn't actually revert anything though
    end
end

-- change player's team etc if they are a cop or not
function policemod_set_player_status(is_a_cop)

    notoriety_reset(PMOD_TEAM_POLICE)
    PMOD_IS_A_COP = is_a_cop

    if not PMOD_IS_A_COP then
        -- not a cop: reset everything
		notoriety_set_max(PMOD_TEAM_POLICE, 5)
		notoriety_restricted_zones_enable(true)
        set_team(LOCAL_PLAYER, PMOD_TEAM_PLAYER)
        team_make_unfriendly(PMOD_TEAM_PLAYER, PMOD_TEAM_POLICE)
        team_make_unfriendly(PMOD_TEAM_POLICE, PMOD_TEAM_PLAYER)
    else
        -- became a cop: enable max notoriety, etc.
        -- note that we don't set the player's team, but instead make the saints and police friendly (in the thread below)
		notoriety_set_max(PMOD_TEAM_POLICE, 0)
        notoriety_restricted_zones_enable(false)
        team_make_allies(PMOD_TEAM_PLAYER, PMOD_TEAM_POLICE)
        team_make_allies(PMOD_TEAM_POLICE, PMOD_TEAM_PLAYER)
        PMOD_CURRENT_NOTORIETY_POINTS = 0
        PMOD_CURRENT_NOTORIETY_DECAY = 0
    end
    
    policemod_set_player_clothes(PMOD_IS_A_COP)
    policemod_update_notoriety_spawns()
end

-- reset player animations to their gender defaults
function policemod_reset_player_anims()
    if character_get_gender(LOCAL_PLAYER) == 0 then
        set_animation_set(LOCAL_PLAYER, "PLYM")
    else
        set_animation_set(LOCAL_PLAYER, "PLYF")
    end
end

-- this should enable/disable tanks etc but doesn't seem to work
function policemod_update_notoriety_spawns()
    notoriety_force_no_spawn(PMOD_TEAM_POLICE, false) -- enable all police spawns?
    if PMOD_IS_A_COP then
        notoriety_force_no_spawn_group("police_tank", not PMOD_SET_TANKS_ENABLED)
        disable_specific_notoriety_group("police_tank", 5, "G_TankGuard", not PMOD_SET_TANKS_ENABLED)
        notoriety_force_no_air_spawn(not PMOD_SET_AIR_ENABLED)
    else
        notoriety_force_no_spawn_group("police_tank", false)
        disable_specific_notoriety_group("police_tank", 5, "G_TankGuard", false)
        notoriety_force_no_air_spawn(false)
    end
end

-- ** threads start **
-- this handles button presses etc.
function policemod_control_thread()
    while true do

        if player_action_is_pressed(PMOD_KEY_MAIN.key) then
            PMOD_HELD_KEY_TICK = PMOD_HELD_KEY_TICK + 1
            if not PMOD_HELD_KEY and PMOD_HELD_KEY_TICK > 75 then
                PMOD_HELD_KEY = true

                if not PMOD_IS_A_COP then
                    policemod_message(PMOD_MSG_CONTROLS_SAINT)
                else
                    policemod_message(PMOD_MSG_CONTROLS_COP)
                end
            end
        elseif PMOD_HELD_KEY then
            PMOD_HELD_KEY = false
            PMOD_HELD_KEY_TICK = 0
        end

        if PMOD_HELD_KEY then
            if player_action_just_pressed(PMOD_KEY_ACTION.key) then
                -- set cop or not and show notification
                policemod_set_player_status(not PMOD_IS_A_COP)
                if PMOD_IS_A_COP then
                    policemod_message(PMOD_MSG_BECAME_COP)
                else
                    policemod_message(PMOD_MSG_BECAME_SAINT)
                end
            elseif PMOD_IS_A_COP and player_action_just_pressed(PMOD_KEY_UP.key) then
                -- increase dispatch level
                local dispatch_level = notoriety_get(PMOD_TEAM_POLICE) + 1

                if dispatch_level == 1 then
                    -- play cellphone action if it's the first call
                    action_play(LOCAL_PLAYER, "cell phone answer")
                end

                if dispatch_level > MAX_NOTORIETY_LEVEL then dispatch_level = MAX_NOTORIETY_LEVEL end
                policemod_update_notoriety_spawns()
                notoriety_set_min_and_max(PMOD_TEAM_POLICE, dispatch_level, dispatch_level)
                PMOD_DISPATCH_TICK = 1
            elseif PMOD_IS_A_COP and player_action_just_pressed(PMOD_KEY_DOWN.key) then
                -- decrease dispatch level
                local dispatch_level = notoriety_get(PMOD_TEAM_POLICE) - 1
                if dispatch_level < 0 then dispatch_level = 0 end
                policemod_update_notoriety_spawns()
                notoriety_set_min_and_max(PMOD_TEAM_POLICE, dispatch_level, dispatch_level)
                PMOD_DISPATCH_TICK = 1
            end
        end

        if PMOD_IS_A_COP and player_action_just_pressed(PMOD_KEY_MAIN.key) then
            -- if F is pressed and they have a hostage
            if character_has_human_shield(LOCAL_PLAYER) then
                local dist_to_npc, npc = get_closest_npc_to_object(LOCAL_PLAYER, 2.0)
                if character_has_specific_human_shield(LOCAL_PLAYER, npc) then
                    -- swap interrogation states
                    PMOD_CURRENT_HOSTAGE = npc
                    policemod_on_hostage_interrogate()
                    -- policemod_message("Hostage: " .. npc)
                end
            end

        elseif PMOD_IS_A_COP and player_action_just_pressed(PMOD_KEY_ACTION.key) then
            -- if E is pressed and they have a hostage
            if PMOD_IS_INTERROGATING and character_has_specific_human_shield(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE) then
                
                -- clean up
                PMOD_IS_INTERROGATING = false
                hud_prompt_clear()

                -- set a bunch of stuff on the detainee
                character_set_only_scripted_grabs(PMOD_CURRENT_HOSTAGE, true)
                player_release_human_shield(LOCAL_PLAYER, true)
                set_ignore_ai_flag(PMOD_CURRENT_HOSTAGE, true)
                character_allow_ragdoll(PMOD_CURRENT_HOSTAGE, false)
                character_prevent_flinching(PMOD_CURRENT_HOSTAGE, true) 
                character_prevent_bumping(PMOD_CURRENT_HOSTAGE, true)
                npc_combat_enable(PMOD_CURRENT_HOSTAGE, false)
                set_dont_attack_me_on_sight_flag(PMOD_CURRENT_HOSTAGE, true)
                npc_go_idle(PMOD_CURRENT_HOSTAGE) 
               
                -- start the arrest sequence
                inv_item_equip(LOCAL_PLAYER, nil)
                policemod_reset_player_anims()
                set_animation_state(PMOD_CURRENT_HOSTAGE, "cower stand variant 1")
                set_animation_set(LOCAL_PLAYER, "COP")

                -- we do this synced thing first to align them, then interrupt it and play the arrest anim
                action_play(PMOD_CURRENT_HOSTAGE, "stand to cower b", "stand to cower b", true, 0.4)
                action_play_synced_non_blocking(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE, "Hostage Release")

                delay(0.3)

                -- do the interrupt
                action_stop_synced_state(LOCAL_PLAYER)
                action_stop_synced_state(PMOD_CURRENT_HOSTAGE)
                ai_clear_scripted_action(LOCAL_PLAYER)
                ai_clear_scripted_action(PMOD_CURRENT_HOSTAGE)
                action_stop_custom(PMOD_CURRENT_HOSTAGE)
                action_stop_custom(LOCAL_PLAYER)
                set_animation_state(PMOD_CURRENT_HOSTAGE, "cower stand variant 1")

                delay(0.05)

                -- now play the arrest anim...
                action_play(LOCAL_PLAYER, "arrest right down", "arrest right down", true, 0.25)
                clear_animation_state(PMOD_CURRENT_HOSTAGE)
                set_animation_state(PMOD_CURRENT_HOSTAGE, "hostage idle")
                policemod_reset_player_anims()

                delay(1.0)

                -- Restore control and add to our list of arrested suspects
                player_human_shield_lock(LOCAL_PLAYER, false)
                PMOD_SUSPECTS[PMOD_CURRENT_HOSTAGE] = true
            end
        end

        -- process this per tick since it handles input
        thread_yield()
    end
end

-- this handles interaction
function policemod_action_thread()
    while true do
        
        if PMOD_IS_A_COP then
            -- if the player is a cop, just routinely set saints/police as allies to avoid any issues
            team_make_allies(PMOD_TEAM_PLAYER, PMOD_TEAM_POLICE)
            team_make_allies(PMOD_TEAM_POLICE, PMOD_TEAM_PLAYER)
            
            local notoriety_level = notoriety_get(PMOD_TEAM_POLICE)

            if PMOD_DISPATCH_TICK >= 1 and notoriety_level > 0 then
                -- Disable the 'min' dispatch notoriety after a while
                PMOD_DISPATCH_TICK = PMOD_DISPATCH_TICK + 1
                if PMOD_DISPATCH_TICK > notoriety_level * 20 then
                    notoriety_set_min(PMOD_TEAM_POLICE, 0)
                    PMOD_DISPATCH_TICK = 0
                end
            end

            -- check for suspects
            if #PMOD_SUSPECTS then
                local dist_to_npc, npc = get_closest_npc_to_object(LOCAL_PLAYER, 1.75)
                if npc and PMOD_SUSPECTS[npc] then
                    if PMOD_CURRENT_SUSPECT ~= npc then
                        -- prompt to release a suspect
                        hud_prompt_clear()
                        hud_prompt(LOCAL_PLAYER, "COP_RELEASE", "policemod_on_hostage_release")
                        PMOD_CURRENT_SUSPECT = npc
                    end
                   
                else
                    if PMOD_CURRENT_SUSPECT then
                        hud_prompt_clear()
                        PMOD_CURRENT_SUSPECT = nil
                    end
                end
            end

            if PMOD_SET_CAN_GO_ROGUE then
                -- update rogue notoriety
                if PMOD_CURRENT_NOTORIETY_DECAY >= 1 then
                    PMOD_CURRENT_NOTORIETY_DECAY = PMOD_CURRENT_NOTORIETY_DECAY - 1
                end

                if PMOD_CURRENT_NOTORIETY_POINTS >= 1 and PMOD_CURRENT_NOTORIETY_DECAY == 0 then
                    PMOD_CURRENT_NOTORIETY_POINTS = PMOD_CURRENT_NOTORIETY_POINTS - 1
                end
                
                if PMOD_CURRENT_NOTORIETY_POINTS >= PMOD_SET_POINTS_TO_GO_ROGUE then
                    policemod_message(PMOD_MSG_GONE_ROGUE)
                    policemod_set_player_status(false)
                    notoriety_set_desired_vehicle_count("Police", 6)
                    notoriety_set_max("Police", 5)
                    notoriety_set("Police", 3) 
                end
            end
        end

        -- we can afford a second between ticks for this one
        delay(1.0)
    end
end

-- ** callbacks start **
function policemod_on_hostage_release()

    if PMOD_CURRENT_SUSPECT and PMOD_SUSPECTS and PMOD_SUSPECTS[PMOD_CURRENT_SUSPECT] then
        -- clear up the prompt
        hud_prompt_clear()

        -- start the release sequence
        clear_animation_state(PMOD_CURRENT_SUSPECT)
        action_play_synced(LOCAL_PLAYER, PMOD_CURRENT_SUSPECT, "Hostage Release")
        
        -- unset properties
        set_ignore_ai_flag(PMOD_CURRENT_SUSPECT, false)
        character_set_only_scripted_grabs(PMOD_CURRENT_SUSPECT, false)
        character_prevent_flinching(PMOD_CURRENT_SUSPECT, false) 
        character_prevent_bumping(PMOD_CURRENT_SUSPECT, false)
        character_allow_ragdoll(PMOD_CURRENT_SUSPECT, true) 

        -- release them to the world so they no longer count in missions etc
        set_cower_variant(PMOD_CURRENT_SUSPECT, get_random_table_entry(PMOD_COWER_VARIANTS))
        set_always_cower_flag(PMOD_CURRENT_SUSPECT, true)
        cower(PMOD_CURRENT_SUSPECT, LOCAL_PLAYER, false, false)
        release_to_world(PMOD_CURRENT_SUSPECT)       
        
        PMOD_SUSPECTS[PMOD_CURRENT_SUSPECT] = nil
        PMOD_CURRENT_SUSPECT = nil
    end

end

function policemod_on_hostage_interrogate()
    if character_has_specific_human_shield(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE) then
        if not PMOD_IS_INTERROGATING then
            -- no ragdolling
            character_allow_ragdoll( LOCAL_PLAYER, false )
            character_allow_ragdoll( PMOD_CURRENT_HOSTAGE, false )

            -- Block controls until the interrogation is complete
            player_human_shield_lock(LOCAL_PLAYER, true)
            
            -- play a synced interrogate animation
	        action_play_synced_state(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE, "Gat Interrogation State")
            action_play_synced(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE, "Gat Interrogation Entry")
            
            -- Add the arrest prompt
            hud_prompt(LOCAL_PLAYER, "COP_ARREST", "policemod_on_hostage_arrest")
            PMOD_IS_INTERROGATING = true
        else
            -- clear any prompt
            hud_prompt_clear()

            -- back to normal
            action_play_synced_state(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE, "HS_Stand")
            action_play_synced(LOCAL_PLAYER, PMOD_CURRENT_HOSTAGE, "Gat Interrogation Return")

            -- Restore control
            player_human_shield_lock(LOCAL_PLAYER, false)
            action_stop_synced_state(LOCAL_PLAYER, true)

            -- Clear invulnerability
            character_allow_ragdoll(LOCAL_PLAYER, true)
            character_allow_ragdoll(PMOD_CURRENT_HOSTAGE, true)
            character_prevent_explosion_fling(PMOD_CURRENT_HOSTAGE, false)
            character_allow_ragdoll(PMOD_CURRENT_HOSTAGE, true)

            PMOD_IS_INTERROGATING = false
        end
    end
end

function policemod_on_hostage_arrest()
    -- we actually just do nothing here and process the keypress itself instead
end

-- tracks crimes by the player
function policemod_on_notoriety_event(notoriety_event_name)
	if PMOD_SET_CAN_GO_ROGUE and PMOD_IS_A_COP and PMOD_NOTORIETY_POINTS[notoriety_event_name] then
        PMOD_CURRENT_NOTORIETY_POINTS = PMOD_CURRENT_NOTORIETY_POINTS + PMOD_NOTORIETY_POINTS[notoriety_event_name]
        PMOD_CURRENT_NOTORIETY_DECAY = PMOD_SET_TIME_BEFORE_POINTS_DECAY
	end
end

-- should make any mission enemy that damages player a priority target for police
function policemod_on_take_damage(victim_name, attacker_name)
	if PMOD_IS_A_COP then
		character_make_priority_target(attacker_name, true)
	end
end

-- ** the end ** =]